
/* module.c: Sony CY8CMBR3116x2 Driver.
 *
 * Copyright 2016 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/miscdevice.h>
#include <linux/input.h>
#include <linux/dma-mapping.h>
#include <linux/earlysuspend.h>
#include <linux/ctype.h>
#include <linux/sysfs.h>
#include <asm/uaccess.h>
#include <mach/mt_gpio.h>
#include <mach/eint.h>
#include <icx_key.h>
#include "cy8cmbr31xx.h"
#include "cy8cmbr3116x2.h"
#include "regs.h"

#define DRV_NAME					"cy8cmbr3116x2"

#if 0
	#define DBGTRACE(fmt, ...)		printk("\e[32m[" DRV_NAME "]       " fmt "\e[0m\n", ## __VA_ARGS__)
	#define ERRTRACE(fmt, ...)		printk("\e[31m[" DRV_NAME "]       " fmt "\e[0m\n", ## __VA_ARGS__)
#else
	#define DBGTRACE(fmt, ...)		//
	#define ERRTRACE(fmt, ...)		printk(      "[" DRV_NAME "] error " fmt      "\n", ## __VA_ARGS__)
#endif


	enum {
		TI2CBOOT					=	15,							/* 起動時間                       */
		TBOOT_SYS_WF				=	1350,						/* システム診断・EMC有効起動時間  */
		TXRES						=	5,							/* 外部リセットパルス幅           */
		TI2C_LATENCY_MAX			=	50,							/* I2C コマンドから実行までの時間 */
		TI2C_LATENCY_SAVE			=	220,						/* SAVE_CHECK_CRC の時間          */
		I2C_SPEED					=	300,						/* I2C 通信速度                   */
		I2C_RETRY					=	10,							/* I2C リトライ回数               */
		CMD_RETRY					=	10,							/* コマンドリトライ回数           */
		CMD_RETRY_INTERVAL			=	50,							/* コマンドリトライ間隔           */
		BOOT_INTERVAL				=	20,							/* 起動待ち間隔                   */
		CY8CMBR3116_ID				=	2565,						/* CY8CMBR3116 デバイス ID        */
		DEBUG_REG_READ				=	50,							/* Read debug reg wait time       */
	} ;

	struct priv ;

	/* デバイス情報 */
	struct deviceinfo {
		struct priv					*p_priv ;						/* プライベートデータ */
		struct cy8cmbr31xx_config	*p_config ;						/* 構成               */
		struct work_struct			work ;							/* ワーク             */
		u8							slave ;							/* スレーブアドレス   */
		int							reset ;							/* リセットポート     */
		u16							mask ;							/* 状態マスク         */
		u16							state ;							/* 以前の状態         */
	} ;

	/* プライベートデータ */
	struct priv {
		struct i2c_client			*p_client ;						/* I2C クライアント */
		struct miscdevice			misc ;							/* MISC デバイス    */
		struct input_dev			*p_input ;						/* 入力デバイス     */
		struct workqueue_struct		*p_workqueue ;					/* ワークキュー     */
		struct timer_list			timer ;							/* タイマ           */
		struct early_suspend		early_suspend ;					/* サスペンド       */
		struct cy8cmbr3116x2_config	*p_config ;						/* 構成             */
		struct deviceinfo			deviceInfo[2] ;					/* デバイス情報     */
		struct class				*sysfs_class;					/* Sysfs            */
		u32							senseor_id ;					/* debug senseor_id */
		u8							init ;							/* 初期化完了フラグ */
		u32							polling ;						/* ポーリング時間   */
	} ;


static struct priv					*m_p_priv ;						/* プライベートデータ */

static int cy8cmbr3116x2_create_sysfs(struct class *sysfs_class);
static void cy8cmbr3116x2_cleanup_sysfs(struct class *sysfs_class);

/* -------------------------------------------------------------------------------- */
/*   キービットを設定                                                               */
/*                                                                                  */
/*   IN : p_input    = 入力デバイス                                                 */
/*        p_key_info = キー情報                                                     */
/*        count      = キー情報数                                                   */
/*   OUT: キービットマスク                                                          */
/* -------------------------------------------------------------------------------- */
static u16 setKeyBit(struct input_dev *p_input, struct cy8cmbr31xx_keyinfo const *p_key_info, uint count)
{
	u16		mask ;
	uint	index ;


	mask = 0 ;

	for(index=0; index<count; index++,p_key_info++) {
		__set_bit(p_key_info->key_code, p_input->keybit) ;
		mask |= p_key_info->key_bit ;
	}

	return mask ;
}


/* -------------------------------------------------------------------------------- */
/*   レジスタを読み込む                                                             */
/*                                                                                  */
/*   IN : p_priv  = プライベートデータ                                              */
/*        slave   = スレーブアドレス                                                */
/*        address = レジスタアドレス                                                */
/*        p_data  = データ格納アドレス                                              */
/*        size    = データサイズ                                                    */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static int read_regs(struct priv *p_priv, u8 slave, u8 address, void *p_data, int size)
{
	struct i2c_msg	msg[2] ;
	int				result ;
	uint			index ;


	memset(&msg[0], 0, sizeof( msg )) ;
	msg[0].addr   = slave ;
	msg[0].flags  = 0 ;
	msg[0].len    = 1 ;
	msg[0].buf    = &address ;
	msg[0].timing = I2C_SPEED ;
	msg[1].addr   = msg[0].addr ;
	msg[1].flags  = I2C_M_RD ;
	msg[1].len    = size ;
	msg[1].buf    = p_data ;
	msg[1].timing = I2C_SPEED ;

	for(index=0; index<I2C_RETRY; index++) {
		result = i2c_transfer(p_priv->p_client->adapter, &msg[0], ARRAY_SIZE( msg )) ;
		if( result == ARRAY_SIZE( msg ) ) {
			return 0 ;
		}
	}

	return -EIO ;
}


/* -------------------------------------------------------------------------------- */
/*   書き込みコマンド完了待ち                                                       */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*        slave  = スレーブアドレス                                                 */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static int wait_cmd(struct priv *p_priv, u8 slave)
{
	struct i2c_msg		msg[2] ;
	u8					address, data ;
	int					result ;
	uint				index ;


	memset(&msg[0], 0, sizeof( msg )) ;
	address       = CTRL_CMD ;
	msg[0].addr   = slave ;
	msg[0].flags  = 0 ;
	msg[0].len    = 1 ;
	msg[0].buf    = &address ;
	msg[0].timing = I2C_SPEED ;
	msg[1].addr   = msg[0].addr ;
	msg[1].flags  = I2C_M_RD ;
	msg[1].len    = 1 ;
	msg[1].buf    = &data ;
	msg[1].timing = I2C_SPEED ;

	for(index=0; index<CMD_RETRY; index++) {
		result = i2c_transfer(p_priv->p_client->adapter, &msg[0], ARRAY_SIZE( msg )) ;
		if( (result==ARRAY_SIZE( msg )) && !data ) {
			return 0 ;
		}

		msleep( CMD_RETRY_INTERVAL ) ;
	}

	return -EIO ;
}


/* -------------------------------------------------------------------------------- */
/*   8bit レジスタを読み込む                                                        */
/*                                                                                  */
/*   IN : p_priv  = プライベートデータ                                              */
/*        slave   = スレーブアドレス                                                */
/*        address = レジスタアドレス                                                */
/*        p_data  = データ格納アドレス                                              */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static inline int read_reg8(struct priv *p_priv, u8 slave, u8 address, u8 *p_data)
{
	u8		data ;
	int		result ;


	result = read_regs(p_priv, slave, address, &data, sizeof( data )) ;
	if( result ) {
		return result ;
	}

	*p_data = data ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   16bit レジスタを読み込む                                                       */
/*                                                                                  */
/*   IN : p_priv  = プライベートデータ                                              */
/*        slave   = スレーブアドレス                                                */
/*        address = レジスタアドレス                                                */
/*        p_data  = データ格納アドレス                                              */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static inline int read_reg16(struct priv *p_priv, u8 slave, u8 address, u16 *p_data)
{
	u8		data[2] ;
	int		result ;


	result = read_regs(p_priv, slave, address, &data[0], sizeof( data )) ;
	if( result ) {
		return -EIO ;
	}

	*p_data = ((u16 )data[1] << 8) | data[0] ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   レジスタへ値を書き込む                                                         */
/*                                                                                  */
/*   IN : p_priv  = プライベートデータ                                              */
/*        slave   = スレーブアドレス                                                */
/*        address = レジスタアドレス                                                */
/*        p_data  = データアドレス                                                  */
/*        size    = データサイズ                                                    */
/*        wait    = コマンド待ち時間（msec）                                        */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/*        -ENOMEM                                                                   */
/* -------------------------------------------------------------------------------- */
static int write_regs(struct priv *p_priv, u8 slave, u8 address, const void *p_data, int size, int wait)
{
	struct i2c_msg	msg ;
	dma_addr_t		phy_addr ;
	u8				buff[8], *p_buff ;
	int				result ;
	uint			index ;

	/* 書き込みコマンド完了待ち */
	result = wait_cmd(p_priv, slave) ;
	if( result ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		return result ;
	}

	/* 初期化 */
	memset(&msg, 0, sizeof( msg )) ;
	msg.addr   = slave ;
	msg.flags  = 0 ;
	msg.len    = 1 + size ;
	msg.buf    = &buff[0] ;
	msg.timing = I2C_SPEED ;
	p_buff     = &buff[0] ;

	/* DMA バッファ確保 */
	if( msg.len > 8 ) {
		p_buff = dma_alloc_coherent(0, msg.len, &phy_addr, GFP_KERNEL) ;
		if( p_buff == NULL ) {
			ERRTRACE("%s(%d)", __func__, __LINE__) ;
			return -ENOMEM ;
		}
		msg.buf      = (unsigned char *)phy_addr ;
		msg.ext_flag = I2C_DMA_FLAG ;
	}

	/* データコピー */
	p_buff[0] = address ;
	memcpy(&p_buff[1], p_data, size) ;

	/* 転送 */
	for(index=0; index<I2C_RETRY; index++) {
		result = i2c_transfer(p_priv->p_client->adapter, &msg, 1) ;
		if( result == 1 ) {
			break ;
		}
	}

	/* DMA バッファ解放 */
	if( p_buff != &buff[0] ) {
		dma_free_coherent(0, 1+size, &p_buff[0], phy_addr) ;
	}

	if( result != 1 ) {
		if( result >= 0 ) {
			result = -EIO ;
		}
		return result ;
	}

	/* コマンド実行待ち */
	if( wait ) {
		msleep( wait ) ;
	}

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   8bit レジスタへ値を書き込む                                                    */
/*                                                                                  */
/*   IN : p_priv  = プライベートデータ                                              */
/*        slave   = スレーブアドレス                                                */
/*        address = レジスタアドレス                                                */
/*        value   = 値                                                              */
/*        wait    = コマンド待ち時間（msec）                                        */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static inline int write_reg8(struct priv *p_priv, u8 slave, u8 address, u8 value, int wait)
{
	return write_regs(p_priv, slave, address, &value, 1, wait) ;
}


/* -------------------------------------------------------------------------------- */
/*   16bit レジスタへ値を書き込む                                                   */
/*                                                                                  */
/*   IN : p_priv  = プライベートデータ                                              */
/*        slave   = スレーブアドレス                                                */
/*        address = レジスタアドレス                                                */
/*        value   = 値                                                              */
/*        wait    = コマンド待ち時間（msec）                                        */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static inline int write_reg16(struct priv *p_priv, u8 slave, u8 address, u16 value, int wait)
{
#ifdef  __LITTLE_ENDIAN
	return write_regs(p_priv, slave, address, &value, 2, wait) ;
#else
	u8	data[2] ;

	data[0] = (u8 )(value >> 0) ;
	data[1] = (u8 )(value >> 8) ;
	return write_regs(p_priv, slave, address, &data[0], 2, wait) ;
#endif/*__LITTLE_ENDIAN__*/
}


/* -------------------------------------------------------------------------------- */
/*   リセット                                                                       */
/*                                                                                  */
/*   IN : p_deviceInfo = デバイス情報                                               */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void reset( struct deviceinfo *p_deviceInfo )
{
	mt_set_gpio_out(p_deviceInfo->reset, GPIO_OUT_ZERO) ;

	msleep( TXRES ) ;

	mt_set_gpio_out(p_deviceInfo->reset, GPIO_OUT_ONE) ;

	msleep( TI2CBOOT ) ;
}


/* -------------------------------------------------------------------------------- */
/*   起動待ち                                                                       */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*        slave  = スレーブアドレス                                                 */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static int wait_boot(struct priv *p_priv, u8 slave)
{
	u16		value ;
	u32		base ;
	int		result ;


	base = jiffies ;

	while( 1 ) {
		result = read_reg16(p_priv, slave, DEVICE_ID, &value) ;
		if( !result ) {
			if( value == CY8CMBR3116_ID ) {
				break ;
			}
			ERRTRACE("%s(%d) : id = %d", __func__, __LINE__, value) ;
		}

		if( (jiffies-base) > msecs_to_jiffies( TBOOT_SYS_WF ) ) {
			return -EIO ;
		}

		DBGTRACE("%s(%d) : retry", __func__, __LINE__) ;

		msleep( BOOT_INTERVAL ) ;
	}

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   フラッシュプログラム                                                           */
/*                                                                                  */
/*   IN : p_deviceInfo = デバイス情報                                               */
/*        p_regs       = レジスタ                                                   */
/*        other_slave  = 他のスレーブアドレス                                       */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static int programFlash(struct deviceinfo *p_deviceInfo, const u8 *p_regs, u8 other_slave)
{
	struct priv		*p_priv=(struct priv *)p_deviceInfo->p_priv ;
	u8				slave ;
	u16				value ;
	int				result ;

	/* リセット */
	reset( p_deviceInfo ) ;

	do {
		/* 起動待ち */
		slave  = p_deviceInfo->slave ;
		result = wait_boot(p_priv, slave) ;
		if( result ) {
			slave  = p_priv->p_client->addr ;
			result = wait_boot(p_priv, slave) ;
			if( result ) {
				slave  = other_slave ;
				result = wait_boot(p_priv, slave) ;
				if( result ) {
					ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
					break ;
				}
			}
		}

		/* コンフィグレーションデータ書き込み */
		result = write_regs(p_priv, slave, DATA_OFFSET, &p_regs[0], 128, TI2C_LATENCY_MAX) ;
		if( result ) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
			break ;
		}

		/* フラッシュ書き込み */
		result = write_reg8(p_priv, slave, CTRL_CMD, SAVE_CHECK_CRC, TI2C_LATENCY_SAVE) ;
		if( result ) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
			break ;
		}

		/* 状態取得 */
		result = read_reg16(p_priv, slave, CTRL_CMD_STATUS, &value) ;
		if( result ) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
			break ;
		}
		if( value & 0xff01 ) {
			ERRTRACE("%s(%d) : 0x%04x", __func__, __LINE__, value) ;
			result = -EIO ;
			break ;
		}

		/* ソフトリセット */
		result = write_reg8(p_priv, slave, CTRL_CMD, SW_RESET, TI2C_LATENCY_MAX) ;
		if( result ) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
			break ;
		}

		/* 変更後にデバイス ID が読めるか確認 */
		result = read_reg16(p_priv, p_deviceInfo->slave, DEVICE_ID, &value) ;
		if( result ) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
			break ;
		}
		if( value != CY8CMBR3116_ID ) {
			ERRTRACE("%s(%d) : 0x%04x", __func__, __LINE__, value) ;
			result = -EIO ;
			break ;
		}

		/* 正常終了 */
		result = 0 ;
	} while( 0 ) ;

	/* リセット状態 */
	mt_set_gpio_out(p_deviceInfo->reset, GPIO_OUT_ZERO) ;

	return result ;
}


/* -------------------------------------------------------------------------------- */
/*   ホスト割り込み処理                                                             */
/*                                                                                  */
/*   IN : なし                                                                      */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void host0_isr( void )
{
	struct priv		*p_priv=m_p_priv ;


	queue_work(p_priv->p_workqueue, &p_priv->deviceInfo[0].work) ;
}

static void host1_isr( void )
{
	struct priv		*p_priv=m_p_priv ;


	queue_work(p_priv->p_workqueue, &p_priv->deviceInfo[1].work) ;
}


/* -------------------------------------------------------------------------------- */
/*   タイマー処理                                                                   */
/*                                                                                  */
/*   IN : data = プライベートデータ                                                 */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void timer_proc( unsigned long data )
{
	struct priv		*p_priv=(struct priv *)data ;


	queue_work(p_priv->p_workqueue, &p_priv->deviceInfo[0].work) ;
	queue_work(p_priv->p_workqueue, &p_priv->deviceInfo[1].work) ;

	p_priv->timer.expires = jiffies + p_priv->polling ;
	add_timer( &p_priv->timer ) ;
}


/* -------------------------------------------------------------------------------- */
/*   ワークキュー処理                                                               */
/*                                                                                  */
/*   IN : p_wok = デバイス情報内のワークアドレス                                    */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void workqueue_proc( struct work_struct *p_work )
{
	struct deviceinfo			*p_deviceInfo=container_of(p_work, struct deviceinfo, work) ;
	struct priv					*p_priv=p_deviceInfo->p_priv ;
	struct cy8cmbr31xx_config	const *p_config=p_deviceInfo->p_config ;
	struct cy8cmbr31xx_keyinfo	const *p_key_info, *p_key_infoEnd ;
	u16							state, update ;
	int							result ;


	do {
		/* ボタン状態取得 */
		result = read_reg16(p_priv, p_deviceInfo->slave, BUTTON_STAT, &state) ;
		if( result ) {
			break ;
		}
		state &= p_deviceInfo->mask ;

		/* 状態に変化がないなら無視 */
		update = state ^ p_deviceInfo->state ;
		if( !update ) {
			break ;
		}
		p_deviceInfo->state = state ;

		/* ボタン状態更新 */
		p_key_info    = &p_config->p_key_info[0] ;
		p_key_infoEnd = &p_config->p_key_info[p_config->count] ;

		do {
			if( update & p_key_info->key_bit ) {
				input_event(p_priv->p_input, EV_KEY, p_key_info->key_code, (state&p_key_info->key_bit)!=0) ;

				update &= ~p_key_info->key_bit ;
				if( !update ) {
					break ;
				}
			}
		} while( ++p_key_info != p_key_infoEnd ) ;

		input_sync( p_priv->p_input ) ;
	} while( 0 ) ;
}


/* -------------------------------------------------------------------------------- */
/*   実行                                                                           */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static int run( struct priv *p_priv )
{
	int		result ;


	do {
		/* リセット */
		reset( &p_priv->deviceInfo[0] ) ;
		reset( &p_priv->deviceInfo[1] ) ;

		/* 起動待ち */
		result = wait_boot(p_priv, p_priv->deviceInfo[0].slave) ;
		if( result ) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
			break ;
		}

		result = wait_boot(p_priv, p_priv->deviceInfo[1].slave) ;
		if( result ) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
			break ;
		}

		p_priv->init = 1 ;

		if( p_priv->deviceInfo[0].p_config->eint_poll >= 0 ) {	/* 割り込み */
			mt_eint_registration(p_priv->deviceInfo[0].p_config->eint_poll, EINTF_TRIGGER_FALLING, host0_isr, 1) ;
			mt_eint_registration(p_priv->deviceInfo[1].p_config->eint_poll, EINTF_TRIGGER_FALLING, host1_isr, 1) ;
			queue_work(p_priv->p_workqueue, &p_priv->deviceInfo[0].work) ;
			queue_work(p_priv->p_workqueue, &p_priv->deviceInfo[1].work) ;
		} else {												/* ポーリング */
			init_timer( &p_priv->timer ) ;
			p_priv->polling        = msecs_to_jiffies( -p_priv->deviceInfo[0].p_config->eint_poll ) ;
			p_priv->timer.expires  = jiffies + p_priv->polling ;
			p_priv->timer.data     = (unsigned long )p_priv ;
			p_priv->timer.function = timer_proc ;
			add_timer( &p_priv->timer ) ;
		}

		/* 正常終了 */
		return 0 ;
	} while( 0 ) ;

	/* リセット状態 */
	mt_set_gpio_out(p_priv->deviceInfo[0].reset, GPIO_OUT_ZERO) ;
	mt_set_gpio_out(p_priv->deviceInfo[1].reset, GPIO_OUT_ZERO) ;

	return result ;
}


/* -------------------------------------------------------------------------------- */
/*   停止                                                                           */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: 0     == 正常終了                                                         */
/*        1     == 既に停止中                                                       */
/* -------------------------------------------------------------------------------- */
static int stop( struct priv *p_priv )
{
	int		eint ;


	if( !p_priv->init ) {
		return 1 ;
	}

	if( p_priv->deviceInfo[0].p_config->eint_poll >= 0 ) {	/* 割り込み */
		eint = p_priv->deviceInfo[0].p_config->eint_poll ;
		mt_eint_mask( eint ) ;
		mt_eint_registration(eint, EINTF_TRIGGER_FALLING, NULL, 0) ;
		mt_eint_mask( eint ) ;

		eint = p_priv->deviceInfo[1].p_config->eint_poll ;
		mt_eint_mask( eint ) ;
		mt_eint_registration(eint, EINTF_TRIGGER_FALLING, NULL, 0) ;
		mt_eint_mask( eint ) ;
	} else {												/* ポーリング */
		del_timer( &p_priv->timer ) ;
	}

	cancel_work_sync( &p_priv->deviceInfo[0].work ) ;
	cancel_work_sync( &p_priv->deviceInfo[1].work ) ;

	/* リセット状態 */
	mt_set_gpio_out(p_priv->deviceInfo[0].reset, GPIO_OUT_ZERO) ;
	mt_set_gpio_out(p_priv->deviceInfo[1].reset, GPIO_OUT_ZERO) ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   開く                                                                           */
/*                                                                                  */
/*   IN : p_inode = i ノード                                                        */
/*        p_file  = ファイル                                                        */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int cy8cmbr3116x2_open(struct inode *p_inode, struct file *p_file)
{
	p_file->private_data = container_of(p_file->private_data, struct priv, misc) ;

	return 0 ;
}


static ssize_t cy8cmbr3116x2_write(struct file *p_file, const char __user *p_buff, size_t size, loff_t *p_pos)
{
	struct priv		*p_priv=p_file->private_data ;
	char			buff[267], *p_srce, *p_dest;
	u8				hi, lo, value, sum ;
	int				result ;

	memset( buff, 0, sizeof( buff ) );

	if( size < sizeof( buff ) ) {
		return -EINVAL ;
	}

	if( copy_from_user(&buff[0], &p_buff[0], sizeof( buff )) ) {
		return -EFAULT ;
	}

	if( memcmp(&buff[0], ":80000000", 9) ) {
		return -EINVAL ;
	}

	/* データとチェックサム取得 */
	p_srce = &buff[9] ;
	p_dest = &buff[0] ;
	sum    = 0x80 ;

	do {
		hi = tolower( p_srce[0] ) ;
		if( !isxdigit( hi ) ) {
			return -EINVAL ;
		}

		lo = tolower( p_srce[1] ) ;
		if( !isxdigit( lo ) ) {
			return -EINVAL ;
		}

		if( hi <= '9' ) {
			value  = (hi -  '0'      ) << 4 ;
		} else {
			value  = (hi - ('a' - 10)) << 4 ;
		}

		if( lo <= '9' ) {
			value |= (lo -  '0'     ) ;
		} else {
			value |= (lo - ('a' - 10)) ;
		}

		*p_dest++ = value ;
		sum      += value ;
		p_srce   += 2 ;
	} while( p_dest != &buff[128+1] ) ;

	if( sum ) {
		return -EINVAL ;
	}

	if( (buff[I2C_ADDR]!=p_priv->deviceInfo[0].slave) && (buff[I2C_ADDR]!=p_priv->deviceInfo[1].slave) ) {
		return -EINVAL ;
	}

	stop( p_priv ) ;

	if( buff[I2C_ADDR] == p_priv->deviceInfo[0].slave ) {
		result = programFlash(&p_priv->deviceInfo[0], (const u8 *)&buff[0], p_priv->deviceInfo[1].slave) ;
	} else {
		result = programFlash(&p_priv->deviceInfo[1], (const u8 *)&buff[0], p_priv->deviceInfo[0].slave) ;
	}
	if( result ) {
		return -EIO ;
	}

	if( p_priv->init ) {
		run( p_priv ) ;
	}

	return size ;
}


/* -------------------------------------------------------------------------------- */
/*   ioctl                                                                          */
/*                                                                                  */
/*   IN : p_file = ファイル                                                         */
/*        cmd    = コマンド                                                         */
/*        arg    = 引数                                                             */
/*   OUT: 0                                                                         */
/*        キーコード数                                                              */
/*        -EFAULT                                                                   */
/*        -EINVAL                                                                   */
/* -------------------------------------------------------------------------------- */
static long cy8cmbr3116x2_ioctl(struct file *p_file, unsigned int cmd, unsigned long arg)
{
	struct priv		*p_priv=p_file->private_data ;


	switch( cmd ) {
		case ICX_KEY_IOCTL_CMD_GETCOUNT:
		{
			return p_priv->p_config->count ;
		}

		case ICX_KEY_IOCTL_CMD_GETINFO:
		{
			if( copy_to_user((void __user *)arg, &p_priv->p_config->p_key_code[0], sizeof( u32 )*p_priv->p_config->count) ) {
				return -EFAULT ;
			}
			break ;
		}

		default:
			return -EINVAL ;
	}

	return 0 ;
}


static const struct file_operations		cy8cmbr3116x2_fops={
											.owner          = THIS_MODULE,
											.open           = cy8cmbr3116x2_open,
											.write          = cy8cmbr3116x2_write,
											.unlocked_ioctl = cy8cmbr3116x2_ioctl,
											.compat_ioctl   = cy8cmbr3116x2_ioctl,
										} ;


/* -------------------------------------------------------------------------------- */
/*   サスペンド処理                                                                 */
/*                                                                                  */
/*   IN : p_early_suspend = サスペンド                                              */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void cy8cmbr3116x2_early_suspend( struct early_suspend *p_early_suspend )
{
	struct priv		*p_priv=container_of(p_early_suspend, struct priv, early_suspend) ;


	DBGTRACE("%s(%d)", __func__, __LINE__) ;

	stop( p_priv ) ;
}


/* -------------------------------------------------------------------------------- */
/*   レジューム処理                                                                 */
/*                                                                                  */
/*   IN : p_early_suspend = サスペンド                                              */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void cy8cmbr3116x2_early_resume( struct early_suspend *p_early_suspend )
{
	struct priv		*p_priv=container_of(p_early_suspend, struct priv, early_suspend) ;


	DBGTRACE("%s(%d)", __func__, __LINE__) ;

	if( p_priv->init ) {
		run( p_priv ) ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   プローブ                                                                       */
/*                                                                                  */
/*   IN : p_client = I2C クライアント                                               */
/*        p_id     = I2C デバイスの識別子                                           */
/*   OUT: 0       == 正常終了                                                       */
/*        0       <  異常終了                                                       */
/* -------------------------------------------------------------------------------- */
static int cy8cmbr3116x2_probe(struct i2c_client *p_client, const struct i2c_device_id *p_id)
{
	struct cy8cmbr3116x2_config		*p_config=p_client->dev.platform_data ;
	struct priv						*p_priv ;
	int								result ;

	/* プライベートデータ確保 */
	p_priv = kzalloc(sizeof( struct priv ), GFP_KERNEL) ;
	if( p_priv == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		return -ENOMEM ;
	}

	/* 初期化 */
	i2c_set_clientdata(p_client, p_priv) ;
	p_priv->p_client               = p_client ;
	p_priv->p_config               = p_config ;
	p_priv->deviceInfo[0].p_priv   = p_priv ;
	p_priv->deviceInfo[0].p_config = &p_config->config[0] ;
	p_priv->deviceInfo[0].slave    = p_config->slave[0] ;
	p_priv->deviceInfo[0].reset    = p_config->config[0].reset ;
	p_priv->deviceInfo[1].p_priv   = p_priv ;
	p_priv->deviceInfo[1].p_config = &p_config->config[1] ;
	p_priv->deviceInfo[1].slave    = p_config->slave[1] ;
	p_priv->deviceInfo[1].reset    = p_config->config[1].reset ;

	/* リセット状態 */
	mt_set_gpio_out(p_priv->deviceInfo[0].reset, GPIO_OUT_ZERO) ;
	mt_set_gpio_out(p_priv->deviceInfo[1].reset, GPIO_OUT_ZERO) ;

	/* MISC デバイス作成 */
	p_priv->misc.minor = MISC_DYNAMIC_MINOR ;
	p_priv->misc.name  = DRV_NAME ;
	p_priv->misc.fops  = &cy8cmbr3116x2_fops ;
	result             = misc_register( &p_priv->misc ) ;
	if( result ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto cy8cmbr3116x2_probe_f00 ;
	}

	/* Create sysfs */
	cy8cmbr3116x2_create_sysfs(p_priv->sysfs_class);

	/* 入力デバイス確保 */
	p_priv->p_input = input_allocate_device() ;
	if( p_priv->p_input == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto cy8cmbr3116x2_probe_e00 ;
	}

	/* 入力デバイス登録 */
	__set_bit(EV_KEY, p_priv->p_input->evbit) ;
	p_priv->deviceInfo[0].mask  = setKeyBit(p_priv->p_input, &p_config->config[0].p_key_info[0], p_config->config[0].count) ;
	p_priv->deviceInfo[1].mask  = setKeyBit(p_priv->p_input, &p_config->config[1].p_key_info[0], p_config->config[1].count) ;
	p_priv->p_input->name       = DRV_NAME ;
	p_priv->p_input->phys       = DRV_NAME ;
	p_priv->p_input->id.bustype = BUS_HOST ;
	p_priv->p_input->id.vendor  = 0x0001 ;
	p_priv->p_input->id.product = 0x0001 ;
	p_priv->p_input->id.version = 0x0100 ;
	p_priv->p_input->dev.parent = &p_client->dev ;
	result                      = input_register_device( p_priv->p_input ) ;
	if( result ) {
		input_free_device( p_priv->p_input ) ;
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		goto cy8cmbr3116x2_probe_e00 ;
	}

	/* ワークキュー作成 */
	p_priv->p_workqueue = create_workqueue( DRV_NAME ) ;
	if( p_priv->p_workqueue == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto cy8cmbr3116x2_probe_d00 ;
	}
	INIT_WORK(&p_priv->deviceInfo[0].work, (void *)workqueue_proc) ;
	INIT_WORK(&p_priv->deviceInfo[1].work, (void *)workqueue_proc) ;

	m_p_priv = p_priv ;

	/* サスペンド登録 */
	p_priv->early_suspend.level   = EARLY_SUSPEND_LEVEL_BLANK_SCREEN + 1 ;
	p_priv->early_suspend.suspend = cy8cmbr3116x2_early_suspend ;
	p_priv->early_suspend.resume  = cy8cmbr3116x2_early_resume ;
	register_early_suspend( &p_priv->early_suspend ) ;

	do {
		/* フラッシュプログラム */
		if( p_config->programFlash ) {

			/* デバイス０ */
			result = programFlash(&p_priv->deviceInfo[0], &p_priv->deviceInfo[0].p_config->p_regs[0], p_config->slave[1]) ;
			if( result ) {
				result = programFlash(&p_priv->deviceInfo[0], &p_priv->deviceInfo[0].p_config->p_regs[0], p_config->slave[1]) ;
				if( result ) {
					ERRTRACE("%s(%d)", __func__, __LINE__) ;
					break ;
				}
			}

			/* デバイス１ */
			result = programFlash(&p_priv->deviceInfo[1], &p_priv->deviceInfo[1].p_config->p_regs[0], p_config->slave[0]) ;
			if( result ) {
				result = programFlash(&p_priv->deviceInfo[1], &p_priv->deviceInfo[1].p_config->p_regs[0], p_config->slave[0]) ;
				if( result ) {
					ERRTRACE("%s(%d)", __func__, __LINE__) ;
					break ;
				}
			}
		}

		/* 実行 */
		run( p_priv ) ;
	} while( 0 ) ;

	return 0 ;


cy8cmbr3116x2_probe_d00:
	input_unregister_device( p_priv->p_input ) ;
cy8cmbr3116x2_probe_e00:
	misc_deregister( &p_priv->misc ) ;
cy8cmbr3116x2_probe_f00:
	kfree( p_priv ) ;

	return result ;
}


/* -------------------------------------------------------------------------------- */
/*   削除                                                                           */
/*                                                                                  */
/*   IN : p_client = I2C クライアント                                               */
/*   OUT: 常に０                                                                    */
/* -------------------------------------------------------------------------------- */
static int cy8cmbr3116x2_remove( struct i2c_client *p_client )
{
	struct priv		*p_priv=i2c_get_clientdata( p_client ) ;


	unregister_early_suspend( &p_priv->early_suspend ) ;

	stop( p_priv ) ;

	destroy_workqueue( p_priv->p_workqueue ) ;

	input_unregister_device( p_priv->p_input ) ;

	/* Cleanup sysfs */
	cy8cmbr3116x2_cleanup_sysfs(p_priv->sysfs_class);

	misc_deregister( &p_priv->misc ) ;

	kfree( p_priv ) ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   Sysfs                                                                          */
/* -------------------------------------------------------------------------------- */
static ssize_t cy8cmbr3116x2_show_debug_raw_count0(struct kobject *kobj,
									struct kobj_attribute *attr, char *buf)
{
	struct priv *p_priv = m_p_priv;
	char temp_buf[32];
	u32  deviceInfo_idx = 0;
	u16  value          = 0;
	int  result         = 0;
	
	deviceInfo_idx = p_priv->senseor_id/16;
	if (deviceInfo_idx < ARRAY_SIZE(p_priv->deviceInfo)) {
		result = read_reg16(p_priv, p_priv->deviceInfo[deviceInfo_idx].slave, DEBUG_RAW_COUNT0, &value);
		if (result) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result);
		} else {
			result = snprintf(temp_buf, sizeof(temp_buf), "%04x\n", value);
			if (result >= 0) {
				memcpy(buf, temp_buf, ((result*2)+1));
			} else {
				result = 0;
			}
		}
	} else {
		return -EINVAL;
	}
	
	return result;
}

static ssize_t cy8cmbr3116x2_show_debug_baseline0(struct kobject *kobj,
									struct kobj_attribute *attr, char *buf)
{
	struct priv *p_priv = m_p_priv;
	char temp_buf[32];
	u32  deviceInfo_idx = 0;
	u16  value          = 0;
	int  result         = 0;
	
	deviceInfo_idx = p_priv->senseor_id/16;
	if (deviceInfo_idx < ARRAY_SIZE(p_priv->deviceInfo)) {
		result = read_reg16(p_priv, p_priv->deviceInfo[deviceInfo_idx].slave, DEBUG_BASELINE0, &value);
		if (result) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result);
		} else {
			result = snprintf(temp_buf, sizeof(temp_buf), "%04x\n", value);
			if (result >= 0) {
				memcpy(buf, temp_buf, ((result*2)+1));
			} else {
				result = 0;
			}
		}
	} else {
		return -EINVAL;
	}
	
	return result;
}

static ssize_t cy8cmbr3116x2_show_debug_senseor_id(struct kobject *kobj,
									struct kobj_attribute *attr, char *buf)
{
	struct priv *p_priv = m_p_priv;
	char temp_buf[32];
	int  result = 0;
	
	result = snprintf( temp_buf, sizeof(temp_buf), "%08x\n", (1<<p_priv->senseor_id) );
	if (result >= 0) {
		memcpy(buf, temp_buf, ((result*2)+1));
	} else {
		result = 0;
	}
	
	return result;
}

static ssize_t cy8cmbr3116x2_store_debug_senseor_id(struct kobject *kobj,
									 struct kobj_attribute *attr,
									 const char *buf, size_t n)
{
	struct priv *p_priv = m_p_priv;
	u32 senseor_state  = 0;
	u32 senseor_id     = 0;
	u32 i              = 0;
	u32 deviceInfo_idx = 0;
	int result         = 0;
	int chk_flag       = 0;
	
	/* param check */
	if (!kobj || !attr || !buf) {
		return -EINVAL;
	}
	
	if( (buf[n-1]=='\n') || (buf[n-1]=='\0') ) {
		senseor_state = simple_strtol(buf, NULL, 16);
		
		for (i=0; i<32; i++) {
			if ((senseor_state >> i) & 1) {
				if (chk_flag) {
					return -EINVAL;
				} else {
					senseor_id = i;
					chk_flag = 1;
				}
			}
		}
		
		deviceInfo_idx = senseor_id/16;
		
		if (deviceInfo_idx < ARRAY_SIZE(p_priv->deviceInfo)) {
			result = write_reg8(p_priv, p_priv->deviceInfo[senseor_id/16].slave, SENSOR_ID, (senseor_id - ((senseor_id/16)*16)), DEBUG_REG_READ);
			if( result ) {
				ERRTRACE("%s(%d) : %d", __func__, __LINE__, result);
			} else {
				p_priv->senseor_id = senseor_id;
			}
		} else {
			return -EINVAL;
		}
	
	} else {
		return -EINVAL;
	}
	
	return n;
}

static ssize_t cy8cmbr3116x2_show_debug_button_stat(struct kobject *kobj,
									struct kobj_attribute *attr, char *buf)
{
	struct priv *p_priv = m_p_priv;
	char temp_buf[32];
	u32  senseor_state = 0;
	int  result        = 0;
	
	senseor_state = ( ((p_priv->deviceInfo[1].state) << 16) | (p_priv->deviceInfo[0].state) );
	
	result = snprintf(temp_buf, sizeof(temp_buf), "%08x\n", senseor_state);
	if (result >= 0) {
		memcpy(buf, temp_buf, ((result*2)+1));
	} else {
		result = 0;
	}
	
	return result;
}

static struct class_attribute cy8cmbr3116x2_debug_button_stat_file = __ATTR(debug_button_stat, 0444, cy8cmbr3116x2_show_debug_button_stat, NULL);
static struct class_attribute cy8cmbr3116x2_debug_raw_count0_file  = __ATTR(debug_raw_count0,  0444, cy8cmbr3116x2_show_debug_raw_count0,  NULL);
static struct class_attribute cy8cmbr3116x2_debug_baseline0_file   = __ATTR(debug_baseline0,   0444, cy8cmbr3116x2_show_debug_baseline0,   NULL);
static struct class_attribute cy8cmbr3116x2_debug_senseor_id_file  = __ATTR(debug_senseor_id,  0666, cy8cmbr3116x2_show_debug_senseor_id,  cy8cmbr3116x2_store_debug_senseor_id);

static int cy8cmbr3116x2_create_sysfs(struct class *sysfs_class)
{
	sysfs_class = class_create(THIS_MODULE, DRV_NAME);
	if ( IS_ERR(sysfs_class) ) {
		return 1;
	}
	
	if ( class_create_file(sysfs_class, &cy8cmbr3116x2_debug_raw_count0_file) ) {
		class_destroy(sysfs_class);
		return 1;
	}
	if ( class_create_file(sysfs_class, &cy8cmbr3116x2_debug_baseline0_file) ) {
		class_destroy(sysfs_class);
		return 1;
	}
	if ( class_create_file(sysfs_class, &cy8cmbr3116x2_debug_senseor_id_file) ) {
		class_destroy(sysfs_class);
		return 1;
	}
	if ( class_create_file(sysfs_class, &cy8cmbr3116x2_debug_button_stat_file) ) {
		class_destroy(sysfs_class);
		return 1;
	}
	
	return 0;
}

static void cy8cmbr3116x2_cleanup_sysfs(struct class *sysfs_class)
{
	class_remove_file(sysfs_class, &cy8cmbr3116x2_debug_button_stat_file);
	class_remove_file(sysfs_class, &cy8cmbr3116x2_debug_senseor_id_file);
	class_remove_file(sysfs_class, &cy8cmbr3116x2_debug_baseline0_file);
	class_remove_file(sysfs_class, &cy8cmbr3116x2_debug_raw_count0_file);
	class_destroy(sysfs_class);
	
	return;
}


static const struct i2c_device_id	cy8cmbr3116x2_id[] = {
										{ DRV_NAME, 0 },
										{}
									} ;

static struct i2c_driver			cy8cmbr3116x2_driver = {
										.probe     = cy8cmbr3116x2_probe,
										.remove    = cy8cmbr3116x2_remove,
										.id_table  = cy8cmbr3116x2_id,
										.driver    = {
											.name  = DRV_NAME,
											.owner = THIS_MODULE,
										},
									} ;


module_i2c_driver( cy8cmbr3116x2_driver ) ;


MODULE_DESCRIPTION("Sony CY8CMBR3116x2 Driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL") ;
MODULE_ALIAS("Sony CY8CMBR3116x2 Driver") ;
MODULE_VERSION( "v0.01" ) ;
